# -*- coding: utf-8 -*-

# Calculate various agreement measures.
# Copyright (C) 2012-2014 Marcin Włodarczak
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division, print_function

import itertools
import re

import numpy as np

from .core import IntervalTier, PointTier, Time

# --------------
# Fleiss's kappa
# --------------


def fleiss_observed_agreement(a):
    '''Return the observed agreement for the input array.'''

    def per_subject_agreement(row):
        '''Return the observed agreement for the i-th subject.'''
        number_of_objects = np.sum(row)
        return (np.sum(np.square(row)) - number_of_objects) / (number_of_objects * (number_of_objects - 1))

    row_probabilities = np.apply_along_axis(per_subject_agreement, axis=1, arr=a)
    return np.mean(row_probabilities)


def fleiss_chance_agreement(a):
    '''Returns the chance agreement for the input array.'''

    def per_category_probabilities(a):
        '''The proportion of all assignments which were to the j-th category.'''
        cat_sums = np.sum(a, axis=0)
        return cat_sums / np.sum(cat_sums)

    return np.sum(np.square(per_category_probabilities(a)))


def fleiss_kappa(a):
    '''Calculates Fleiss's kappa for the input array (with categories
    in columns and items in rows).'''

    p = fleiss_observed_agreement(a)
    p_e = fleiss_chance_agreement(a)
    return (p - p_e) / (1 - p_e)

# -------------
# Cohen's kappa
# -------------


def cohen_kappa(a):
    '''Calculates Cohen's kappa for the input array.'''

    totsum = np.sum(a)
    colsums = np.sum(a, 0)
    rowsums = np.sum(a, 1)
    # Observed agreement.
    p = np.sum(np.diagonal(a)) / totsum
    # Chance agreement.
    p_e = np.sum((colsums * rowsums) / totsum ** 2)
    return (p - p_e) / (1 - p_e)

# ----------
# Scott's pi
# ----------


def scott_pi(a):
    '''Calculates Scott's Pi for the input array.'''

    totsum = np.sum(a)
    colsums = np.sum(a, 0)
    rowsums = np.sum(a, 1)
    # Observed agreement.
    p = np.sum(np.diagonal(a)) / totsum
    # Chance agreement.
    joint_marginal_props = (colsums + rowsums) / (2 * totsum)
    p_e = np.sum(joint_marginal_props ** 2)
    return (p - p_e) / (1 - p_e)

# ----------------------------------------------------------
# Functions producing contingency tables from lists of labels
# ----------------------------------------------------------

def align_labels(tiers_list, precision=None, regex=r'[^\s]+'):
    '''Create a list of lists for all time-aligned Interval
    or Point object in tiers_list, whose text matches regex.
    For example:

    [[label_1-tier_1, label_1-tier_2, label_1-tier_3],
     [label_2-tier_1, label_2-tier_2, label_2-tier_3],
     ...
     [label_n-tier_n, label_n-tier_n, label_n-tier_n]]

    The allowed mismatch between object's timestamps can be
    controlled via the precision parameter.
    '''

    # If precision is supplied, temporarily change
    # the value of Time._precision
    if precision is not None:
        precision_old = Time._precision
        Time._precision = precision

    if len(tiers_list) < 2:
        raise Exception('At least two tiers need to be provided')
    # Check if all elements of tiers_list are either
    # either PointTiers or IntervalTiers.
    elif (not (all([isinstance(x, IntervalTier) for x in tiers_list])
               or all([isinstance(x, PointTier) for x in tiers_list]))):
        raise TypeError('Only objects of types IntervalTier or PointTier can be aligned.')
    elif len(set([len(x) for x in tiers_list])) > 1:
        raise Exception('Input tiers differ in the number of objects.')

    labels_aligned = []
    for intervals in zip(*[x for x in tiers_list]):
        start_times = [x.start_time for x in intervals]
        end_times = [x.end_time for x in intervals]
        labels = [x.text.strip() for x in intervals]
        if any([not re.search(regex, x) for x in labels]):
            # Only go on if labels of all intervals match the regular expression.
            continue
        # Check if start and end times match.
        elif start_times.count(start_times[0]) != len(start_times):
            raise Exception('Objects\' time stamps do not match: {0}'.format(start_times))
        elif end_times.count(end_times[0]) != len(end_times):
            raise Exception('Objects\' time stamps do not match: {0}'.format(end_times))
        else:
            labels_aligned.append(labels)

    # Reset Time._precision to its original value
    if precision is not None:
        Time._precision = precision_old

    return labels_aligned

def cont_table(tiers_list, precision, regex):
    '''Produce a contingency table from annotations in tiers_list
    whose text matches regex, and whose time stamps are not
    misaligned by more than precision.
    '''

    labels_aligned = align_labels(tiers_list, precision, regex)

    # List of unique labels from both lists.
    categories = list(set(itertools.chain(*labels_aligned)))

    # A 2-by-2 array
    if len(labels_aligned[0]) == 2:
        categories_product = itertools.product(categories, categories)
        table = np.array([labels_aligned.count(list(x)) for x in categories_product])
        table.shape = (len(categories), len(categories))
    # An n-by-m array
    else:
        table = np.array([x.count(y) for x in labels_aligned for y in categories])
        table.shape = (len(labels_aligned), len(categories))
    return table

def agreement(tiers_list, method, precision=None, regex=r'[^\s]+'):

    _AGREEMENT_METHODS = {'cohen-kappa': cohen_kappa,
                          'fleiss-kappa': fleiss_kappa,
                          'scott-pi': scott_pi}

    if method not in _AGREEMENT_METHODS:
        available_methods = ', '.join(_AGREEMENT_METHODS.keys())
        raise Exception('Unsupported method. Available options are {0}.'.format(available_methods))
    elif len(tiers_list) < 2:
        raise Exception('At least two tiers need to be provided')
    elif len(tiers_list) == 2 and method in ['cohen-kappa', 'scott-pi']:
        agr = _AGREEMENT_METHODS[method](cont_table(tiers_list, precision, regex))
        return [x.name for x in tiers_list] + [agr]
    elif len(tiers_list) > 2 and method == 'fleiss-kappa':
        agr = _AGREEMENT_METHODS[method](cont_table(tiers_list, precision, regex))
        return [x.name for x in tiers_list] + [agr]
    else:
        tier_combinations = itertools.combinations(tiers_list, 2)
        agr = []
        for tiers_pair in tier_combinations:
            agr_pair = _AGREEMENT_METHODS[method](cont_table(tiers_pair, precision, regex))
            agr.append([x.name for x in tiers_pair] + [agr_pair])
        return agr



# -*- coding: utf-8 -*-

# TextGridTools -- Read, write, and manipulate Praat TextGrid files
# Copyright (C) 2011-2014 Hendrik Buschmeier, Marcin Włodarczak
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division

import bisect
import warnings
import copy
import math
import operator
import re


__all__ = [
    'TextGrid',
    'Tier', 'IntervalTier', 'PointTier',
    'Annotation', 'Interval', 'Point',
    'Time'
]


class TextGrid(object):
    '''A TextGrid.'''

    def __init__(self, filename=''):
        super(TextGrid, self).__init__()
        self._tiers = []
        self.filename = filename

    tiers = property(fget=lambda self: self._tiers,
                     doc='Tiers in this TextGrid object.')

    def add_tier(self, tier):
        '''Add a tier.'''
        self._tiers.append(tier)

    def add_tiers(self, tiers):
        '''Add a list of tiers.'''
        for tier in tiers:
            self.add_tier(tier)

    def insert_tier(self, tier, position):
        '''Insert a tier at the specified position.'''
        self._tiers.insert(position, tier)

    def delete_tier(self, tier_name):
        '''Delete a tier.'''
        tier = self.get_tier_by_name(tier_name)
        self.tiers.remove(tier)

    def delete_tiers(self, tier_names, complement=False):
        '''Delete a list of tiers.

        If complement is False, delete tiers with the specified names.
        If complement is True, delete tiers not specified.
        '''
        if not complement:
            self._tiers = [tier for tier in self._tiers 
                if tier.name not in tier_names]
        else:
            self._tiers = [tier for tier in self._tiers
                if tier.name in tier_names]

    def get_tier_names(self):
        '''Get names of all tiers.'''
        return [tier.name for tier in self._tiers]

    def has_tier(self, name):
        '''Check whether TextGrid has a tier of the specified name.'''
        return name in self.get_tier_names()

    def get_tier_by_name(self, name):
        '''Get tier of specified name.'''
        for tier in self._tiers:
            if tier.name == name:
                return tier
        raise ValueError('Textgrid ' + self.filename +
                    ' does not have a tier called "' + name + '".')

    def _get_earliest_start_time(self):
        '''Return the earliest start time among all tiers.'''
        return min([tier.start_time for tier in self])

    start_time = property(fget=_get_earliest_start_time,
                          doc='TextGrid start time.')

    def _get_latest_end_time(self):
        '''Return the latest end time among all tiers.'''
        return max([tier.end_time for tier in self])

    end_time = property(fget=_get_latest_end_time,
                        doc='TextGrid end time.')

    def __contains__(self, name):
        return self.has_tier(name)

    def __iter__(self):
        '''Return an iterator over the tiers of this TextGrid.'''
        return iter(self._tiers)

    def __len__(self):
        '''Return the number of tiers.'''
        return len(self._tiers)


class Tier(object):
    "An abstract tier."

    def __init__(self, start_time=0, end_time=0, name='', objects=None):
        super(Tier, self).__init__()
        self._specd_start_time = Time(start_time)
        self._specd_end_time = Time(end_time)
        self.name = name
        self._objects = []
        if objects is not None and objects != []:
            self.add_annotations(objects)

    def _get_start_time(self):
        '''Get start time of this tier.'''
        if len(self._objects) > 0:
            return min([self._objects[0].start_time, self._specd_start_time])
        else:
            return self._specd_start_time

    start_time = property(fget=_get_start_time, doc='Start time.')

    def _get_end_time(self):
        '''Get end time of this tier.'''
        if len(self._objects) > 0:
            return max([self._objects[-1].end_time, self._specd_end_time])
        else:
            return self._specd_end_time

    end_time = property(fget=_get_end_time, doc='End time.')

    def add_annotation(self, obj):
        '''Adds an annotation object to this tier.

        The annotation object is inserted at the correct position within the
        tier. If the space is already (partially) occupied by a different
        annotation object, a ValueError is raised.
        '''
        if ((len(self._objects) > 0 and obj.start_time >= self._objects[-1].end_time) 
                or len(self._objects) == 0): # can we simply append obj?
            self._objects.append(obj)
        else: # no, we need to insert it
            overlapping_objects = self.get_annotations_between_timepoints(
                obj.start_time, obj.end_time, 
                left_overlap=True, right_overlap=True)
            if overlapping_objects == []:
                start_timepoints = [interval.start_time for interval in self._objects]
                position = bisect.bisect_left(start_timepoints, obj.start_time)
                self._objects.insert(position, obj)
            else:
                raise ValueError(
                    'Could not add object {0} to this tier: Overlap.'.format(
                        repr(obj)))

    def add_annotations(self, objects):
        '''Add a sequence of annotation objects.'''
        for obj in objects:
            self.add_annotation(obj)

    def _get_annotations(self):
        '''Get all intervals of this tier.'''
        return self._objects

    annotations = property(fget=_get_annotations,
                doc='The list of annotations of this tier.')

    def _get_annotation_index_by_start_time(self, time):
        '''Get annotation index of the object that starts at time.'''
        idx = bisect.bisect_left([obj.start_time for obj in self], time)
        if (idx < len(self) and self._objects[idx].start_time == time):
            return idx
        else:
            return None

    def _get_annotation_index_by_end_time(self, time):
        '''Get the annotation index of the object that ends at time.'''
        idx = bisect.bisect_left([obj.end_time for obj in self], time)
        if (idx < len(self) and self._objects[idx].end_time == time):
            return idx
        else:
            return None

    def _get_annotation_indices_by_time(self, time):
        '''Get annotation indices at the specified time.'''
        idx = bisect.bisect_left([obj.end_time for obj in self], time)
        if (idx < len(self._objects) and 
            time >= self._objects[idx].start_time):
            if (len(self._objects) > idx+1
                and self._objects[idx+1].start_time == time):
                return [idx, idx+1]
            else:
                return [idx]
        else:
            return []

    def _get_annotation_index_range_between_timepoints(self, start, end, left_overlap=False, right_overlap=False):
        '''Get annotation index range for objects between start and end.

        If left_overlap or right_overlap is False annotation objects
        overlapping with start or end are excluded.
        '''
        start_timepoints = [interval.start_time for interval in self._objects]
        end_timepoints = [interval.end_time for interval in self._objects]
        if left_overlap:
            index_lo = bisect.bisect_right(end_timepoints, start)
        else:
            index_lo = bisect.bisect_left(start_timepoints, start)
        if right_overlap:
            index_hi = bisect.bisect_left(start_timepoints, end)
        else:
            index_hi = bisect.bisect_right(end_timepoints, end)
        return (index_lo, index_hi)

    def get_annotation_by_start_time(self, time):
        '''Get the annotation object that starts at time.'''
        idx = self._get_annotation_index_by_start_time(time)
        return idx if idx is None else self._objects[idx]

    def get_annotation_by_end_time(self, time):
        '''Get the annotation object that ends at time.'''
        idx = self._get_annotation_index_by_end_time(time)
        return idx if idx is None else self._objects[idx]

    def get_annotations_by_time(self, time):
        '''Get annotation objects at the specified time.'''
        indices = self._get_annotation_indices_by_time(time)
        return [self._objects[idx] for idx in indices]

    def get_annotations_between_timepoints(self, start, end, left_overlap=False, right_overlap=False):
        '''Get annotation objects between start and end.

        If left_overlap or right_overlap is False annotation objects
        overlapping with start or end are excluded.
        '''
        idx_lo, idx_hi = self._get_annotation_index_range_between_timepoints(start, end, left_overlap, right_overlap)
        return self._objects[idx_lo:idx_hi]

    def get_nearest_annotation(self, time, pattern=r'.*', boundary='both',
                               direction='both', exclude_overlapped=False):
        '''Get a list of the annotation object(s) nearest to time.

        Boundary specifies whether the distance to an annotation object
        is calculated based on its start time ('start'), end time
        ('end') or both ('both'). Direction specifies whether it is
        looked to the left hand side of time ('left'), to the right
        hand side of time ('right') or to both sides ('both').
        Annotations overlapping with time can be excluded.

        Note: When time lies exactly on a boundary, this boundary is
        both to the left and to the right of time.
        '''
        # Filter for specified regular expression
        matching_objects = self.get_annotations_with_matching_text(
            pattern=pattern, regex=True)
        # Exclude overlapping intervals from search
        if exclude_overlapped:
            overlapping_objects = self.get_annotations_by_time(time)
            for oo in overlapping_objects:
                if oo in matching_objects:
                    matching_objects.remove(oo)
        # Extract start and end boundary times and calculate their
        # distance to the reference point, this gives a list of tuples
        # (boundary time, distance, start/end)
        if boundary in ['start', 'both']:
            start_boundaries = [(obj.start_time, time - obj.start_time,
                'start') for obj in matching_objects]
        if boundary in ['end', 'both']:
            end_boundaries = [(obj.end_time, time - obj.end_time,
                'end') for obj in matching_objects]
        # Depending on search direction and boundary type, compute the
        # candidate boundary times with minimum distance to the
        # reference point. Candidates are still tuples
        candidates = []
        if direction in ['left', 'both']:
            if boundary in ['start', 'both']:
                # Filter for left hand side
                sl = [x for x in start_boundaries if x[1] >= 0]
                if len(sl) > 0:
                    candidates.append(min(sl, key=operator.itemgetter(1)))
            if boundary in ['end', 'both']:
                # Filter for left hand side
                el = [x for x in end_boundaries if x[1] >= 0]
                if len(el) > 0:
                    candidates.append(min(el, key=operator.itemgetter(1)))
        if direction in ['right', 'both']:
            if boundary in ['start', 'both']:
                # Filter for right hand side, make distance positive
                sr = [(x[0], x[1]*-1, x[2]) for x in start_boundaries if x[1] <= 0]
                if len(sr) > 0:
                    candidates.append(min(sr, key=operator.itemgetter(1)))
            if boundary in ['end', 'both']:
                # Filter for right hand side, make distance positive
                er = [(x[0], x[1]*-1, x[2]) for x in end_boundaries if x[1] <= 0]
                if len(er) > 0:
                    candidates.append(min(er, key=operator.itemgetter(1)))
        # Compute corresponding annotation objects for all candidates
        # that have the minimum distance to the reference point and
        # collect the unique ones
        if len(candidates) > 0:
            min_distance = min(candidates, key=operator.itemgetter(1))[1]
            results = set()
            for candidate in candidates:
                if candidate[1] == min_distance:
                    if candidate[2] == 'start':
                        results.add(self.get_annotation_by_start_time(candidate[0]))
                    elif candidate[2] == 'end':
                        results.add(self.get_annotation_by_end_time(candidate[0]))
            return sorted(list(results), key=lambda x: x.start_time)
        else:
            return list()

    def get_annotations_with_text(self, pattern='', n=0, regex=False):
        '''Get annotation objects with text matching the pattern.

        If n > 0 the first n matches are returned, if n < 0, the last
        n matches are returned, if n = 0 all matches are returned. The 
        pattern is treated as a regular expression, if regex is True.
        '''
        if regex:
            result = [obj for obj in self if re.search(pattern, obj.text)]
        else:
            result = [obj for obj in self if obj.text == pattern]
        if n == 0:
            return result  # Return all matching intervals
        elif n > 0:
            return result[:n]  # Return the first n matching intervals
        else:  # i.e., n < 0
            return result[n:]  # Return the last n matching intervals

    def get_annotations_with_matching_text(self, pattern='', n=0, regex=False):
        '''Get annotation objects with text matching the pattern.

        If n > 0 the first n matches are returned, if n < 0, the last
        n matches are returned, if n = 0 all matches are returned. The 
        pattern is treated as a regular expression, if regex is True.

        Note: get_annotations_with_matching_text is deprecated. Use
        get_annotations_with_text instead.
        '''
        warnings.warn('get_annotations_with_matching_text is deprecated. '\
                      'Use get_annotations_with_text instead.',DeprecationWarning)
        return self.get_annotations_with_text(pattern, n, regex)

    def delete_annotation_by_start_time(self, time):
        '''Delete the annotation object that starts at time.'''
        idx = self._get_annotation_index_by_start_time(time)
        if idx is not None:
            del self._objects[idx]

    def delete_annotation_by_end_time(self, time):
        '''Delete the annotation object that ends at time.'''
        idx = self._get_annotation_index_by_end_time(time)
        if idx is not None:
            del self._objects[idx]

    def delete_annotations_by_time(self, time):
        '''Delete annotation objects at the specified time.'''
        indices = self._get_annotation_indices_by_time(time)
        for idx in reversed(indices): # Needs to be done in reverse order
            del self._objects[idx]

    def delete_annotations_between_timepoints(self, start, end, left_overlap=False, right_overlap=False):
        '''Delete annotation objects between start and end.

        If left_overlap or right_overlap is False annotation objects
        overlapping with start or end are excluded.
        '''
        idx_lo, idx_hi = self._get_annotation_index_range_between_timepoints(start, end, left_overlap, right_overlap)
        r = range(idx_hi-1,idx_lo-1,-1)
        for idx in r: # Needs to be done in reverse order
            del self._objects[idx]

    def delete_annotations_with_text(self, pattern='', n=0, regex=False):
        '''Delete annotation objects with text matching the pattern.

        If n > 0 the first n matches are deleted, if n < 0, the last
        n matches are deleted, if n = 0 all matches are deleted. The 
        pattern is treated as a regular expression, if regex is True.
        '''
        # Inefficient implementation, but works
        annotations = self.get_annotations_with_text(pattern, n, regex)
        for annotation in annotations:
            self.delete_annotation_by_start_time(annotation.start_time)

    def delete_empty_annotations(self):
        '''Delete annotation object with empty or whitespace-only text.
        '''
        self.delete_annotations_with_text(pattern=r'^\s*$', regex=True)

    def tier_type(self):
        '''Return the type of the tier as a string.'''
        return self.__class__.__name__
       
    def __iter__(self):
        return iter(self._objects)

    def __getitem__(self, key):
        return self._objects[key]

    def __delitem__(self, key):
        del self._objects[key]

    def __len__(self):
        '''Return number of annotation objects in this tier.'''
        return len(self._objects)

    def __repr__(self):
        return '{0}(start_time={1}, end_time={2}, name="{3}", objects={4})'.format(self.__class__.__name__,
            self.start_time, self.end_time, self.name, self._objects)


class IntervalTier(Tier):
    '''An IntervalTier.'''

    def __init__(self, start_time=0, end_time=0, name='', objects=None):
        super(IntervalTier, self).__init__(Time(start_time), Time(end_time),
            name, objects)

    def add_interval(self, interval):
        '''Add an interval to this tier.'''
        self.add_annotation(interval)

    def add_intervals(self, intervals):
        '''Add a sequence of intervals to this tier.'''
        self.add_annotations    (intervals)

    def _get_intervals(self):
        '''Get all intervals of this tier.'''
        return self._objects

    intervals = property(fget=_get_intervals,
                doc='The list of intervals of this tier.')

    def get_copy_with_gaps_filled(self, start_time=None, end_time=None, empty_string=''):
        '''Returns a copy where gaps are filled with empty intervals.'''
        tier_copy = copy.deepcopy(self)
        if start_time is not None:
            tier_copy._specd_start_time = Time(start_time)
        if end_time is not None:
            tier_copy._specd_end_time = Time(end_time)
        # If no intervals exist, add one interval from start to end
        if len(self) == 0:
            empty = Interval(self.start_time, self.end_time, empty_string)
            tier_copy.add_annotation(empty)
        else:
            # If necessary, add empty interval at start of tier
            if self.intervals[0].start_time > tier_copy.start_time:
                empty = Interval(tier_copy.start_time, self.intervals[0].start_time, empty_string)
                tier_copy.add_annotation(empty)
            # If necessary, add empty interval at end of tier
            if self.intervals[-1].end_time < tier_copy.end_time:
                empty = Interval(self.intervals[-1].end_time, tier_copy.end_time, empty_string)
                tier_copy.add_annotation(empty)
            # Insert empty intervals in between non-meeting intervals
            for i in range(len(self) - 1):
                if self.intervals[i].end_time < self.intervals[i+1].start_time:
                    empty = Interval(self.intervals[i].end_time, self.intervals[i+1].start_time, empty_string)
                    tier_copy.add_annotation(empty)
        return tier_copy

    def get_copy_with_same_intervals_merged(self):
        '''Returns a copy of TIER in which consecutive intervals
        with identical labels are merged.'''

        res = IntervalTier(name=self.name)
        prev_intr = None

        for intr in self:
            if prev_intr is None:
                prev_intr = copy.copy(intr)
            elif prev_intr.text != intr.text or prev_intr.end_time != intr.start_time:
                res.add_interval(prev_intr)
                prev_intr = copy.copy(intr)
            else:
                prev_intr.end_time = intr.end_time
        else:
            if prev_intr is not None:
                res.add_interval(prev_intr)
        return res


class PointTier(Tier):
    '''A PointTier (also "TextTier").'''

    def __init__(self, start_time=0, end_time=0, name='', objects=None):
        super(PointTier, self).__init__(Time(start_time), Time(end_time),
            name, objects)

    def add_point(self, point):
        '''Add a point to this tier.'''
        self.add_annotation(point)

    def add_points(self, points):
        '''Adds a list of points to this tier.'''
        self.add_annotations(points)

    def _get_points(self):
        '''Get all points of this tier.'''
        return self._objects

    points = property(fget=_get_points,
                doc='The list of points of this tier.')

    def tier_type(self):
        '''Return the type of the tier as a string. 

        As Praat's point tiers are for some reason called "TextTier" in the 
        TextGrid file format we simply return a string literal here.
        '''
        return 'TextTier'


class Annotation(object):

    def __init__(self, start_time, end_time, text=''):
        '''Initialise the Annotation.'''
        super(Annotation, self).__init__()
        self._start_time = Time(start_time)
        self._end_time = Time(end_time)
        if start_time > end_time:
            raise ValueError('Start time {0} after end time {1}.'.format(start_time, end_time))
        self.text = text.strip()

    def _get_start_time(self):
        return self._start_time

    def _set_start_time(self, start_time):
        if start_time > self.end_time:
            raise ValueError('Start time {0} after end time {1}.'.format(start_time, self.end_time))
        self._start_time = Time(start_time)

    start_time = property(fget=_get_start_time, fset=_set_start_time,
        doc='The start time.')

    def _get_end_time(self):
        return self._end_time

    def _set_end_time(self, end_time):
        if end_time < self.start_time:
            raise ValueError('Start time {0} after end time {1}.'.format(self.start_time, end_time))
        self._end_time = Time(end_time)

    end_time = property(fget=_get_end_time, fset=_set_end_time,
        doc='The end time.')

    def duration(self):
        '''Get duration of this annotation.'''
        return self.end_time - self.start_time

    def __eq__(self, other):
        return (type(self) == type(other)
                and self.start_time == other.start_time
                and self.end_time == other.end_time
                and self.text == other.text)

    def __ne__(self, other):
        return not self.__eq__(other)

    __hash__ = object.__hash__

    def __repr__(self):
        return u'Annotation({0}, {1}, "{2}")'.format(self.start_time,
            self.end_time, self.text)


class Interval(Annotation):
    '''An interval of two points of time with a text label.'''

    def __init__(self, start_time, end_time, text=''):
        '''Initialise this Interval.'''
        super(Interval, self).__init__(start_time, end_time, text)

    def __repr__(self):
        return u'Interval({0}, {1}, "{2}")'.format(self.start_time, self.end_time, self.text)


class Point(Annotation):
    '''A point of time with a text label.

    Internally an Annotation where start time equals end time.
    '''

    def __init__(self, time, text=''):
        '''Initialise this Point.'''
        super(Point, self).__init__(time, time, text)

    def _get_time(self):
        '''Return time, i.e., start_time.'''
        return self._start_time

    def _set_time(self, time):
        '''Set time, i.e., start time and end time.'''
        self._start_time = self._end_time = Time(time)

    time = start_time = end_time = property(fget=_get_time, fset=_set_time,
        doc='The point of time.')

    def __repr__(self):
        return u'Point({0}, "{1}")'.format(self.time, self.text)


class Time(float):
    '''A representation of point in time with a predefined precision.'''

    _precision = 0.0001

    def __eq__(self, other):
        return math.fabs(self - other) < self._precision

    def __ne__(self, other):
        return math.fabs(self - other) >= self._precision

    __hash__ = float.__hash__

    def __gt__(self, other):
        return self != other and self - other > 0

    def __lt__(self, other):
        return self != other and self - other < 0

    def __ge__(self, other):
        return self == other or self > other

    def __le__(self, other):
        return self == other or self < other



# -*- coding: utf-8 -*-

# TextGridTools -- Read, write, and manipulate Praat TextGrid files
# Copyright (C) 2011-2014 Hendrik Buschmeier, Marcin Włodarczak
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from .core import TextGrid
from .core import Tier, IntervalTier, PointTier
from .core import Annotation, Interval, Point
from .core import Time

if sys.version_info < (3, 0):
	from . import io
	from .io import read_textgrid, read_eaf, write_to_file
else:
	from . import io3 as io
	from .io3 import read_textgrid, read_eaf, write_to_file

from . import agreement, util

__all__ = [
    'TextGrid',
    'Tier', 'IntervalTier', 'PointTier',
    'Annotation', 'Interval', 'Point',
    'Time',
    'read_textgrid', 'read_eaf', 'write_to_file',
    'agreement', 'io', 'util',

]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# print-tiernames.py - Prints the tier names of a textrid file
# Copyright (C) 2012 Hendrik Buschmeier
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Prints the name of the tiers of the specified textgrid files.
# If a directory is specified, all tier names of the .TextGrid
# files contained in this directory are printed.

from __future__ import print_function, division

import os
import sys

import tgt

EXTENSION = 'TextGrid'

def print_tiernames(filenames):
	for filename in filenames:
		try:
			tg = tgt.io.read_textgrid(filename)
			print(filename)
			for tiername in tg.get_tier_names():
				print('\t' + tiername)
		except Exception, err:
			print(filename + ' caused a problem.')
			sys.stderr.write('ERROR: %s\n' % str(err))


def main(argv=None):
	if argv is None:
		argv = sys.argv
	if len(argv) < 2:
		print('USAGE: ' + argv[0] + ' <list of TextGrid files or directories>')
	list_of_tg_files = []
	for arg in argv[1:]:
		if os.path.exists(arg):
			if os.path.isdir(arg):
				list_of_tg_files += [arg + '/' + filename for filename in os.listdir(arg) if filename.endswith(EXTENSION)]
			else:
				list_of_tg_files.append(arg)
		else:
			raise FileNotFoundException(arg)
	print_tiernames(list_of_tg_files)

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Concatenate Praat TextGrid files.
# Copyright (C) 2011 Marcin Włodarczak
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse

import tgt

def parse_arguments():
    """Parse the command-line arguments."""
    argparser = argparse.ArgumentParser(description='Concatenate Praat TextGrid files.')
    argparser.add_argument('-i', type=str, nargs='+', required=True, dest='input_files',
                        help='Space-separated list of paths to TextGrid files in the\
                              order they should be concatenated.')
    argparser.add_argument('-o', type=str, nargs=1, required=True, dest='output_file',
                        help='Path to the resulting file.')
    argparser.add_argument('-e', type=str, nargs=1, dest='encoding', default='utf-8',
                        help='File encoding (defaults to UTF-8).')
    argparser.add_argument('-t', type=str, nargs=1, dest='type', default='short',
                        help='TextGrid format: short, long (defaults to short)')
    return vars(argparser.parse_args())

def main():
    # Parse the command-line arguments.
    args = parse_arguments()
    # Read all TextGrids into a list.
    textgrids = [tgt.io.read_textgrid(path, args['encoding']) for path in args['input_files']]
    textgrid_concatenated = tgt.util.concatenate_textgrids(textgrids)
    # Write the concatenated TextGrid to a file.
    tgt.io.write_to_file(textgrid=textgrid_concatenated, filename=args['output_file'][0], format='short', encoding=args['encoding'])


if __name__ == '__main__':
    main()





# TextGridTools -- Read, write, and manipulate Praat TextGrid files
# Copyright (C) 2011-2014 Hendrik Buschmeier, Marcin Włodarczak
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import datetime
import collections
import xml.etree.ElementTree as ET

from .core import TextGrid, IntervalTier, Interval, PointTier, Point, Time


def read_textgrid(filename, encoding='utf-8', include_empty_intervals=False):
    '''Read a Praat TextGrid file and return a TextGrid object. 
    If include_empty_intervals is False (the default), empty intervals
    are excluded. If True, they are included. Empty intervals from specific
    tiers can be also included by specifying tier names as a string (for one tier)
    or as a list.'''
    with open(filename, 'rt', encoding=encoding) as f:
        # Read whole file into memory ignoring empty lines and lines consisting
        # solely of a single double quotes.
        stg = [line.strip() for line in f.readlines()
            if line.strip() not in ['', '"']]
    if stg[0] != 'File type = "ooTextFile"':
        raise Exception(filename)
    if stg[1] != 'Object class = "TextGrid"':
        raise Exception(filename)
    # Determine the TextGrid format.
    if stg[2].startswith('xmin'):
        return read_long_textgrid(filename, stg, include_empty_intervals)
    else:
        return read_short_textgrid(filename, stg, include_empty_intervals)


def read_short_textgrid(filename, stg, include_empty_intervals=False):
    '''Read a Praat short TextGrid file and return a TextGrid object.'''

    def read_interval_tier(stg_extract):
        '''Read and return an IntervalTier from a short TextGrid.'''
        name = stg_extract[1].strip('"')  # name w/o quotes
        start_time = Time(stg_extract[2])
        end_time = Time(stg_extract[3])
        include_empty_intervals_this_tier = include_empty_intervals_in_tier(name, include_empty_intervals)
        it = IntervalTier(start_time, end_time, name)
        i = 5
        while i < len(stg_extract):
            text = stg_extract[i + 2].strip('"') # text w/o quotes
            if text.strip() != '' or include_empty_intervals_this_tier:
                it.add_annotation(Interval(
                    Time(stg_extract[i]), # left bound
                    Time(stg_extract[i + 1]), # right bound
                    text))
            i += 3               
        return it

    def read_point_tier(stg_extract):
        '''Read and return a PointTier (called TextTier) from a short TextGrid.'''
        name = stg_extract[1].strip('"')  # name w/o quotes
        start_time = stg_extract[2]
        end_time = stg_extract[3]
        pt = PointTier(start_time, end_time, name)
        i = 5
        while i < len(stg_extract):
            text = stg_extract[i + 1].strip('"') # text w/o quotes
            pt.add_annotation(Point(
                stg_extract[i], # time
                text))   
            i += 2
        return pt

    tg = TextGrid(filename)
    read_start_time = stg[2]
    read_end_time = stg[3]
    if stg[4] != '<exists>':
        raise Exception(filename)
    read_no_of_tiers = stg[5]
    index = 6
    while index < len(stg):
        num_obj = int(stg[index + 4])
        if stg[index] == '"IntervalTier"':
            tg.add_tier(read_interval_tier(stg[index:index + 5 + num_obj * 3]))
            index += 5 + (num_obj * 3)
        elif stg[index] == '"TextTier"':
            tg.add_tier(read_point_tier(stg[index:index + 5 + num_obj * 2]))
            index += 5 + (num_obj * 2)
        else:
            raise Exception('Unknown tier type: {0}'.format(stg[index]))
    return tg


def read_long_textgrid(filename, stg, include_empty_intervals=False):
    '''Read a Praat long TextGrid file and return a TextGrid object.'''

    def get_attr_val(x):
        """Extract the attribute value from a long TextGrid line."""
        return x.split(' = ')[1]

    def read_interval_tier(stg_extract):
        '''Read and return an IntervalTier from a long TextGrid.'''
        name = get_attr_val(stg_extract[2])[1:-1]  # name w/o quotes
        start_time = get_attr_val(stg_extract[3])
        end_time = get_attr_val(stg_extract[4])
        include_empty_intervals_this_tier = include_empty_intervals_in_tier(name, include_empty_intervals)
        it = IntervalTier(start_time, end_time, name)
        i = 7
        while i < len(stg_extract):
            text = get_attr_val(stg_extract[i + 2])[1:-1] # text w/o quotes
            if text.strip() != '' or include_empty_intervals_this_tier:
                it.add_annotation(Interval(
                    Time(get_attr_val(stg_extract[i])), # left bound
                    Time(get_attr_val(stg_extract[i + 1])), # right bound
                    text))
            i += 4
        return it

    def read_point_tier(stg_extract):
        '''Read and return a PointTier (called TextTier) from a long TextGrid.'''
        name = get_attr_val(stg_extract[2])[1:-1]  # name w/o quotes
        start_time = get_attr_val(stg_extract[3])
        end_time = get_attr_val(stg_extract[4])
        pt = PointTier(start_time, end_time, name)
        i = 7
        while i < len(stg_extract):
            text = get_attr_val(stg_extract[i + 1])[1:-1] # text w/o quotes
            pt.add_annotation(Point(
                Time(get_attr_val(stg_extract[i])), # time
                text))
            i += 3
        return pt

    tg = TextGrid(filename)
    read_start_time = get_attr_val(stg[2])
    read_end_time = get_attr_val(stg[3])
    if stg[4].split()[1] != '<exists>':
        raise Exception(filename)
    read_no_of_tiers = get_attr_val(stg[5])
    index = 7
    while index < len(stg):
        num_obj = int(get_attr_val(stg[index + 5]))
        if get_attr_val(stg[index + 1]) == '"IntervalTier"':
            tg.add_tier(read_interval_tier(stg[index:index + 6 + num_obj * 4]))
            index += 6 + (num_obj * 4)
        elif get_attr_val(stg[index + 1]) == '"TextTier"':
            tg.add_tier(read_point_tier(stg[index:index + 6 + num_obj * 3]))
            index += 6 + (num_obj * 3)
        else:
            raise Exception('Unknown tier type: {0}'.format(stg[index]))
    return tg

def include_empty_intervals_in_tier(tier_name, include_empty_intervals):
    """Check whether to include empty intervals for a particular tier"""
    if isinstance(include_empty_intervals, bool):
        return include_empty_intervals
    elif isinstance(include_empty_intervals, str):
        return tier_name == include_empty_intervals
    elif isinstance(include_empty_intervals, list):
        return tier_name in include_empty_intervals
    else:
        raise TypeError('Invalid type of include_empty_intervals: {0}.'.format(type(include_empty_intervals)))


def read_eaf(filename):
    eaf_tree = ET.parse(filename)

    time_units = eaf_tree.find('HEADER').attrib['TIME_UNITS']
    if time_units != 'milliseconds':
        raise Exception('Unsupported TIME_UNITS format: {}. Only "milliseconds" are supported.'.format(time_units))

    tg = TextGrid(filename=filename)
    time_slots = dict((ts.attrib['TIME_SLOT_ID'], float(ts.attrib['TIME_VALUE']) / 1000)
                      for ts in eaf_tree.iterfind('TIME_ORDER/TIME_SLOT'))

    for eaf_tr in eaf_tree.iterfind('TIER'):
        tgt_tier = IntervalTier(name=eaf_tr.attrib['TIER_ID'])

        for intr in eaf_tr.iterfind('ANNOTATION/'):
            if intr.tag == 'ALIGNABLE_ANNOTATION':
                tgt_tier.add_interval(Interval(
                    start_time=time_slots[intr.attrib['TIME_SLOT_REF1']],
                    end_time=time_slots[intr.attrib['TIME_SLOT_REF2']],
                    text=intr.find('ANNOTATION_VALUE').text))
            else:
                raise Exception('Only ALIGNABLE_ANNOTATIONs are supported at the moment')
        tg.add_tier(tgt_tier)
    return tg

##  Functions for writing TextGrid files
##----------------------------------------------------------------------------

def correct_start_end_times_and_fill_gaps(textgrid):
    '''Correct the start/end times of all tiers and fill gaps.

    Returns a copy of a textgrid, where empty gaps between intervals
    are filled with empty intervals and where start and end times are
    unified with the start and end times of the whole textgrid.
    '''
    textgrid_copy = copy.deepcopy(textgrid)
    for tier in textgrid_copy:
        if isinstance(tier, IntervalTier):
            tier_corrected = tier.get_copy_with_gaps_filled(textgrid.start_time, textgrid.end_time)
            position = textgrid_copy.tiers.index(tier)
            textgrid_copy.tiers[position] = tier_corrected
    return textgrid_copy

def export_to_short_textgrid(textgrid):
    '''Convert a TextGrid object into a string of Praat short TextGrid format.'''
    result = ['File type = "ooTextFile"',
              'Object class = "TextGrid"',
              '',
              str(textgrid.start_time),
              str(textgrid.end_time),
              '<exists>',
              str(len(textgrid))]
    textgrid_corrected = correct_start_end_times_and_fill_gaps(textgrid)
    for tier in textgrid_corrected:
        result += ['"' + tier.tier_type() + '"',
                   '"' + tier.name + '"',
                   str(tier.start_time), str(tier.end_time), str(len(tier))]
        if isinstance(tier, IntervalTier):
            result += [u'{0}\n{1}\n"{2}"'.format(obj.start_time, obj.end_time, obj.text)
                       for obj in tier]
        elif isinstance(tier, PointTier):
            result += [u'{0}\n"{1}"'.format(obj.time, obj.text)
                       for obj in tier]
        else:
            raise Exception('Unknown tier type: {0}'.format(tier.name))
    return '\n'.join(result)


def export_to_long_textgrid(textgrid):
    """Convert a TextGrid object into a string of Praat long TextGrid format."""
    result = ['File type = "ooTextFile"',
              'Object class = "TextGrid"',
              '',
              'xmin = ' + str(textgrid.start_time),
              'xmax = ' + str(textgrid.end_time),
              'tiers? <exists>',
              'size = ' + str(len(textgrid)),
              'item []:']
    textgrid_corrected = correct_start_end_times_and_fill_gaps(textgrid)
    for i, tier in enumerate(textgrid_corrected):
        result += ['\titem [{0}]:'.format(i + 1),
                   '\t\tclass = "{0}"'.format(tier.tier_type()),
                   '\t\tname = "{0}"'.format(tier.name),
                   '\t\txmin = ' + str(tier.start_time),
                   '\t\txmax = ' + str(tier.end_time),
                   '\t\tintervals: size = ' + str(len(tier))]
        if isinstance(tier, IntervalTier):
            for j, obj in enumerate(tier):
                result += ['\t\tintervals [{0}]:'.format(j + 1),
                           '\t\t\txmin = ' + str(obj.start_time),
                           '\t\t\txmax = ' + str(obj.end_time),
                           '\t\t\ttext = "' + obj.text + '"']
        elif isinstance(tier, PointTier):
            for j, obj in enumerate(tier):
                result += ['\t\tpoints [{0}]:'.format(j + 1),
                           '\t\t\tnumber = ' + str(obj.time),
                           '\t\t\tmark = "' + obj.text + '"']
        else:
            raise Exception('Unknown tier type: {0}'.format(tier.name))
    return '\n'.join(result)


def export_to_elan(textgrid, encoding='utf-8', include_empty_intervals=False,
                   include_point_tiers=True, point_tier_annotation_duration=0.04):
    """Convert a TextGrid object into a string of ELAN eaf format."""

    time_slots = collections.OrderedDict()

    def get_time_slot_id(time, ts_dict=time_slots):
        """Returns (and possibly creates) the time slot id of time."""
        time_in_ms = int(time * 1000)
        if time_in_ms not in ts_dict:
            ts_id = 'ts' + str(len(ts_dict) + 1)
            ts_dict[time_in_ms] = ts_id
        return ts_dict[time_in_ms]

    # Create ELAN header
    head = [
        u'<?xml version="1.0" encoding="{0}"?>'.format(encoding.upper()),
        u'<ANNOTATION_DOCUMENT AUTHOR="TextGridTools" DATE="{0}" FORMAT="2.7" VERSION="2.7" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://www.mpi.nl/tools/elan/EAFv2.7.xsd">'.format(datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S+00:00')),
        u'<HEADER MEDIA_FILE="" TIME_UNITS="milliseconds">',
        u'\t<PROPERTY NAME="lastUsedAnnotationId">0</PROPERTY>',
        u'</HEADER>']
    # Create annotations
    annotation_id_count = 1
    annotations = []
    textgrid_copy = correct_start_end_times_and_fill_gaps(textgrid) if include_empty_intervals else textgrid
    for tier in textgrid_copy:
        annotations.append(u'<TIER DEFAULT_LOCALE="en" LINGUISTIC_TYPE_REF="default-lt" TIER_ID="{0}">'.format(tier.name))
        if isinstance(tier, IntervalTier):
            for interval in tier:
                if not include_empty_intervals and interval.text == '':
                    continue
                annotations += [
                    u'<ANNOTATION>',
                    u'\t<ALIGNABLE_ANNOTATION ANNOTATION_ID="{0}" TIME_SLOT_REF1="{1}" TIME_SLOT_REF2="{2}">'.format('a' + str(annotation_id_count), get_time_slot_id(interval.start_time), get_time_slot_id(interval.end_time)),
                    u'\t\t<ANNOTATION_VALUE>{0}</ANNOTATION_VALUE>'.format(interval.text),
                    u'\t</ALIGNABLE_ANNOTATION>',
                    u'</ANNOTATION>']
                annotation_id_count += 1
        elif isinstance(tier, PointTier):
            if include_point_tiers:
                for point in tier.points:
                    annotations += [
                        u'<ANNOTATION>',
                        u'\t<ALIGNABLE_ANNOTATION ANNOTATION_ID="{0}" TIME_SLOT_REF1="{1}" TIME_SLOT_REF2="{2}">'.format('a' + str(annotation_id_count), get_time_slot_id(point.time), get_time_slot_id(point.time + point_tier_annotation_duration)),
                        u'\t\t<ANNOTATION_VALUE>{0}</ANNOTATION_VALUE>'.format(point.text),
                        u'\t</ALIGNABLE_ANNOTATION>',
                        u'</ANNOTATION>']
                    annotation_id_count += 1
        else:
            raise Exception('Unknown tier type: {0}'.format(tier.name))
        annotations.append(u'</TIER>')
    # Create time stamp information
    time_info = [u'<TIME_ORDER>']
    for time_value, time_slot_id in time_slots.items():
        time_info.append(u'\t<TIME_SLOT TIME_SLOT_ID="{0}" TIME_VALUE="{1}"/>'.format(time_slot_id, str(time_value)))
    time_info.append(u'</TIME_ORDER>')
    # Create ELAN footer
    foot = [u'<LINGUISTIC_TYPE GRAPHIC_REFERENCES="false" LINGUISTIC_TYPE_ID="default-lt" TIME_ALIGNABLE="true"/>',
            u'<LOCALE COUNTRY_CODE="US" LANGUAGE_CODE="en"/>',
            u'</ANNOTATION_DOCUMENT>']
    eaf = head + time_info + annotations + foot
    return '\n'.join(eaf)


def export_to_table(textgrid, separator=','):
    """Convert a TextGrid object into a table with fields delimited
    with the specified separator (comma by default)."""
    result = [separator.join(['tier_name', 'tier_type', 'start_time', 'end_time', 'text'])]

    for tier in textgrid:
        if isinstance(tier, IntervalTier):
            for obj in tier:
                if obj.text:
                    result.append(separator.join([tier.name, tier.tier_type(),
                                                  str(obj.start_time), str(obj.end_time), obj.text]))
        elif isinstance(tier, PointTier):
            for obj in tier:
                result.append(separator.join([tier.name, tier.tier_type(),
                                              str(obj.time), str(obj.time), obj.text]))
        else:
            raise Exception('Unknown tier type: {0}'.format(tier.name))
    return '\n'.join(result)


# Listing of currently supported export formats.
_EXPORT_FORMATS = {
    # Export to Praat TextGrid in short format
    'short': export_to_short_textgrid,
    # Export to Praat TextGrid in long (i.e., standard) format
    'long': export_to_long_textgrid,
    # Export to ELAN .eaf format
    'eaf': export_to_elan,
    # Export to a table
    'table': export_to_table
}


def write_to_file(textgrid, filename, format='short', encoding='utf-8', **kwargs):
    """Write a TextGrid object to a file in the specified format."""
    with open(filename, 'w', encoding=encoding) as f:
        if format in _EXPORT_FORMATS:
            f.write(_EXPORT_FORMATS[format](textgrid, **kwargs))
        else:
            raise Exception('Unknown output format: {0}'.format(format))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# TextGridTools -- Read, write, and manipulate Praat TextGrid files
# Copyright (C) 2011-2014 Hendrik Buschmeier
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from distutils.core import setup

setup(
    name='tgt',
    description='TextGridTools -- Read, write, and manipulate Praat TextGrid files',
    version='1.3.0',
    packages=['tgt'],
    scripts=[
    	'scripts/tgt-concatenate-textgrids.py',
        'tgt-extract-part.py',
    	'scripts/tgt-print-tiernames.py',
    ],
    maintainer='Hendrik Buschmeier',
    maintainer_email='hbuschme@uni-bielefeld.de',
    license='GNU General Public License 3',
    download_url='http://github.com/hbuschme/TextGridTools/',
    classifiers=[
    	'Development Status :: 5 - Production/Stable',
    	'Intended Audience :: Science/Research',
    	'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    	'Programming Language :: Python',
    	'Programming Language :: Python :: 2.7',
    	'Programming Language :: Python :: 3',
    	'Programming Language :: Python :: 3.4',
    	'Topic :: Scientific/Engineering :: Information Analysis',
    ],
)



# -*- coding: utf-8 -*-

# TextGridTools -- Read, write, and manipulate Praat TextGrid files
# Copyright (C) 2011-2014 Hendrik Buschmeier, Marcin Włodarczak
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division

import re
import collections
import copy

from .core import TextGrid, IntervalTier, Interval

##  High-level functions
##----------------------------------------------------------------------------

def shift_boundaries(tier, left, right):
    """
    Return a copy of the tier with boundaries shifted by the specified
    amount of time (in seconds). Positive values expand the tier and negative
    values shrink it, i.e.:
    * positive value of left shifts the left boundary to the left
    * negative value of left shifts the left boundary to the right
    * positive value of right shifts the right boundary to the right
    * negative value of right shifts the right boundary to the left.
    """

    tier_end_shifted = tier.end_time + left + right
    tier_shifted = IntervalTier(start_time=0,
                                end_time=tier_end_shifted,
                                name=tier.name)

    for i, interval in enumerate(tier.intervals):

        if interval.end_time <= left * -1:
            continue

        if interval.start_time > left * -1:
            interval_start_shifted = interval.start_time + left
        else:
            interval_start_shifted = 0

        interval_end_shifted = interval.end_time + left
        if (interval_start_shifted >= tier_end_shifted):
            break
        elif interval_end_shifted > tier_end_shifted:
            interval_end_shifted = tier_end_shifted

        tier_shifted.add_annotation(Interval(interval_start_shifted,
                                             interval_end_shifted,
                                             interval.text))
    return tier_shifted


def get_overlapping_intervals(tier1, tier2, regex=r'[^\s]+', overlap_label=None):
    '''Return a list of overlaps between intervals of tier1 and tier2.
    If no overlap_label is specified, concatenated labels
    of overlapping intervals are used as the resulting label.
    
    All nonempty intervals are included in the search by default.
    '''
    if not isinstance(tier2, IntervalTier):
        raise TypeError('Argument is not an IntervalTier')
    intervals1 = tier1.intervals
    intervals2 = tier2.intervals
    overlaps = []
    i, j = 0, 0
    while i < len(tier1) and j < len(tier2):
        lo = max(intervals1[i].start_time, intervals2[j].start_time)
        hi = min(intervals1[i].end_time, intervals2[j].end_time)
        if (lo < hi and re.search(regex, intervals1[i].text)
            and re.search(regex, intervals2[j].text)):
            if overlap_label is None:
                text = '+'.join([intervals1[i].text, intervals2[j].text])
            else:
                text = overlap_label
            overlaps.append(Interval(lo, hi, text))
        if intervals1[i].end_time < intervals2[j].end_time:
            i += 1
        else:
            j += 1
    return overlaps


def concatenate_textgrids(textgrids, ignore_nonmatching_tiers=False):
    '''Concatenate Tiers with matching names.

    TextGrids are concatenated in the order they are specified. If 
    ignore_nonmatching_tiers is False, an exception is raised if the
    number and the names of tiers differ between TextGrids.
    '''
    tier_names_intersection = set.intersection(
        *[set(tg.get_tier_names()) for tg in textgrids])
    # Check whether the TextGrids have the same number of tiers
    # and whether tier names match. If they don't
    # and if ignore_nonmatching_tiers is False, raise an exception.
    if (not ignore_nonmatching_tiers
        and not all([len(tier_names_intersection) == len(tg) for tg in textgrids])):
        raise Exception('TextGrids have different numbers of tiers or tier names do not match.')
    tot_duration = 0
    tiers = {}  # tier_name : tgt.Tier()
    for textgrid in textgrids:
        for tier in textgrid:
            if tier.name not in tier_names_intersection:
                continue
            intervals = []
            # If this is the first we see this tier, we just make a copy
            # of it as it is.
            if tier.name not in tiers.keys():
                tiers[tier.name] = copy.deepcopy(tier)
            # Otherwise we update the start and end times of intervals
            # and append them to the first part.
            else:
                for interval in tier.intervals:
                    interval.start_time += tot_duration
                    interval.end_time += tot_duration
                    intervals.append(interval)
                tiers[tier.name].add_annotations(intervals)
        tot_duration += textgrid.end_time
    # Create a new TextGrid and add the concatenated tiers
    textgrid_concatenated = TextGrid()
    # Add tiers in the order they're found in the first TextGrid.
    textgrid_concatenated.add_tiers(
        [tiers[x] for x in tier_names_intersection])
    return textgrid_concatenated


def merge_textgrids(textgrids, ignore_duplicates=True):
    '''Return a TextGrid object with tiers in all textgrids.p

    If ignore_duplicates is False, tiers with equal names are renamed
    by adding a path of the textgrid or a unique number incremented
    with each occurrence.
    '''
    tg_merged = TextGrid()
    tier_duplicates_lookup = collections.defaultdict(int)
    for tg in textgrids:
        for tier in tg:
            tier_copy = copy.deepcopy(tier)
            if tg_merged.has_tier(tier.name):
                if not ignore_duplicates:
                    if tg.filename.strip() != '':
                        tier_copy.name += '_' + tg.filename
                    else:
                        tier_duplicates_lookup[tier.name] += 1
                        tier_copy.name += '_' + str(tier_duplicates_lookup[tier.name])
                else:
                    continue
            tg_merged.add_tier(tier_copy)
    return tg_merged

## Chronogram
## ---------------------------------------------------------------------------

def chronogram(tiers, speech_label=None, silence_label=None):

    '''Construct a chronogram between intervals in input tiers.
    Interval labels are classified as silences or volcalisations
    by matching them against the speech_label and the silence_label
    regular expressions. By default all intervals with empty or
    whitespace-only labels are treated as silences.

    The code is a generalisation of Jaffe and Feldstein's (1970) 
    6-state Markov model to an arbitrary number of speakers. Instances
    of silences and overlaps are classified as within-speaker-overlap
    (wso), between-speaker-overlap (bso), within-speaker-silence (wss)
    or between-speaker-silence (bss).  Individual vocalistions are
    labelled with the the source tier name.
    '''
    

    # Calculate communicatfive states for each tier.
    communicative_states = classify_communicative_state(tiers, speech_label, silence_label)
    
    is_joint_state = lambda st: st == 'none' or st.find(',') > 0
    is_single_state = lambda st: not is_joint_state(st)

    chrono = IntervalTier(name='chronogram-{0}'.format('-'.join(t.name for t in tiers)))
    prev_single = None

    for i in range(len(communicative_states)):

        cur_start = communicative_states[i].start_time
        cur_end = communicative_states[i].end_time
        cur_state = communicative_states[i].text

        # Make sure there are no consecutive same state sequences
        if i > 0:
            prev_state = communicative_states[i - 1].text
            assert  prev_state != cur_state,\
                'Consecutive same states: {0}, {1}'.format(prev_state, cur_state)

        if is_joint_state(cur_state):

            # If we have not seen a single-speaker vocalisation, skip it.
            if  prev_single is None:
                continue

            try:
                next_state = communicative_states[i + 1].text
            except IndexError:
                next_state = None
            
            # Transitions between joint states do not result in speaker change.
            # The same is true for transitions from a joint state to a single state 
            # equal to the previous single state and for file-final joint states.
            if (next_state is None or is_joint_state(next_state)
                or (is_single_state(next_state) and prev_single == next_state)):

                chrono.add_interval(
                    Interval(start_time=cur_start, end_time=cur_end,
                             text='wso:{{{}}}'.format(cur_state) if cur_state != 'none' else 'wss'))
            else:
                chrono.add_interval(
                    Interval(start_time=cur_start, end_time=cur_end,
                             text='bso:{{{}}}'.format(cur_state) if cur_state != 'none' else 'bss'))
        # Label single vocalisations with the source tier name.
        elif is_single_state(cur_state):
            chrono.add_interval(Interval(start_time=cur_start, end_time=cur_end, text=cur_state))
            prev_single = cur_state
        else:
            raise Exception('Unknown cummunicative state: {0}'.format(cur_state))

    # FIXME: Should a dictionary with communicative labels and the chronogram be 
    # returned instead of the chronogram itself?
    # return {'communicative_labels': communicative_states, 'chronogram': chrono}
    return chrono

def communicative_labels(tiers, voc_re=None, silence_re=None):

    if silence_re is not None:
        speech_tiers = [t.name for t in tiers if re.search(silence_re, t[0].text) is None]
    else:
        if speech_re is None:
            speech_re = r'[^\s]+',
        speech_tiers = [t.name for t in tiers if re.search(speech_re, t[0].text) is not None]

    if not speech_tiers:
        return 'none'
    else:
        return ','.join(speech_tiers)

def classify_communicative_state(tiers, speech_label=None, silence_label=None):

    # Fill all gaps with empty intervals and ensure the tiers have
    # identical start and end times
    start_time_earliest = min(tier.start_time for tier in tiers)
    end_time_latest = max(tier.end_time for tier in tiers)

    tiers_filled = [tier.get_copy_with_gaps_filled(start_time=start_time_earliest,
                                                   end_time=end_time_latest)
                    for tier in tiers]

    communicative_states = IntervalTier(name='communicative_states')


    while all(tiers_filled):
        lo = max(x[0].start_time for x in tiers_filled)
        hi = min(x[0].end_time for x in tiers_filled)
        
        if lo < hi:
            com_state = communicative_labels(tiers_filled, speech_label, silence_label)
            communicative_states.add_annotation(Interval(lo, hi, com_state))

        for t in tiers_filled:
            if t[0].end_time == hi:
                del t[0]

    # Merge consecutive intervals with indentical labels
    communicative_states = communicative_states.get_copy_with_same_intervals_merged()
    return communicative_states
